# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from .holidays import Locale


class da_DK(Locale):
    """
    01-01: [NF] Nytårsdag
    06-05: [NF] Grundlovsdag
    12-25: [NRF] Juledag
    12-26: [NRF] Anden juledag
    3 days before Easter: [NRV] Skærtorsdag
    2 days before Easter: [NRV] Langfredag
    Easter: [NRV] Påskedag
    1 day after Easter: [NRV] Anden påskedag
    26 days after Easter: [NRV] Store bededag
    39 days after Easter: [NRV] Kristi himmelfartsdag
    49 days after Easter: [NRV] Pinsedag
    50 days after Easter: [NRV] Anden pinsedag
    """

    locale = "da-DK"
    easter_type = EASTER_WESTERN



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Holiday, Locale


class en_NZ(Locale):
    """
    01-01: [NF] New Year's Day
    02-06: [NF] Waitangi Day
    04-25: [NF] ANZAC Day
    12-25: [NRF] Christmas Day
    12-26: [NF] Boxing Day
    2 days before Easter: [NRV] Good Friday
    1 day after Easter: [NRV] Easter Monday
    1. monday in june: [NV] Queen's Birthday
    4. monday in october: [NV] Labour Day
    """

    locale = "en-NZ"
    easter_type = EASTER_WESTERN

    def holiday_new_years_day_observed(self):
        date = SmartDayArrow(self.year, 1, 1)

        if date.weekday() in ['saturday', 'sunday']:
            return [Holiday(
                    locale=self.locale,
                    region="",
                    date=date.shift_to_weekday('monday', including=True),
                    description="New Year's Day (observed)",
                    flags="NV",
                    notes="")]

        return []

    def holiday_after_new_years_day(self):
        date = SmartDayArrow(self.year, 1, 2)

        if date.weekday() in ['sunday', 'monday']:
            return [Holiday(
                locale=self.locale,
                region="",
                date=date.shift_to_weekday('tuesday', including=True),
                description="Day after New Year's Day",
                flags="NV",
                notes="")]

        elif date.weekday() == 'saturday':
            return [Holiday(
                locale=self.locale,
                region="",
                date=date.shift_to_weekday('monday', including=True),
                description="Day after New Year's Day",
                flags="NV",
                notes="")]

        return [Holiday(
                locale=self.locale,
                region="",
                date=date,
                description="Day after New Year's Day",
                flags="NV",
                notes="")]

    def holiday_waitangi_day_observed(self):
        date = SmartDayArrow(self.year, 2, 6)

        if self.year > 2016 and date.weekday() in ['saturday', 'sunday']:
            return [Holiday(
                    locale=self.locale,
                    region="",
                    date=date.shift_to_weekday('monday', including=True),
                    description="Waitangi Day (observed)",
                    flags="NV",
                    notes="")]

        return []

    def holiday_anzac_day_observed(self):
        date = SmartDayArrow(self.year, 4, 25)

        if self.year > 2015 and date.weekday() in ['saturday', 'sunday']:
            return [Holiday(
                    locale=self.locale,
                    region="",
                    date=date.shift_to_weekday('monday', including=True),
                    description="ANZAC Day (observed)",
                    flags="NV",
                    notes="")]

        return []

    def holiday_christmas_day_observed(self):
        date = SmartDayArrow(self.year, 12, 25)

        if date.weekday() == 'sunday':
            return [Holiday(
                locale=self.locale,
                region="",
                date=date.shift_to_weekday('tuesday', including=True),
                description="Christmas Day (observed)",
                flags="NV",
                notes="")]

        elif date.weekday() == 'saturday':
            return [Holiday(
                locale=self.locale,
                region="",
                date=date.shift_to_weekday('monday', including=True),
                description="Christmas Day (observed)",
                flags="NV",
                notes="")]

        return []

    def holiday_boxing_day_observed(self):
        date = SmartDayArrow(self.year, 12, 26)

        if date.weekday() == 'sunday':
            return [Holiday(
                locale=self.locale,
                region="",
                date=date.shift_to_weekday('tuesday', including=True),
                description="Boxing Day (observed)",
                flags="NV",
                notes="")]

        elif date.weekday() == 'saturday':
            return [Holiday(
                locale=self.locale,
                region="",
                date=date.shift_to_weekday('monday', including=True),
                description="Boxing Day (observed)",
                flags="NV",
                notes="")]

        return []



# coding=utf-8
from holidata.utils import SmartDayArrow
from .holidays import Locale, Holiday

"""
sources:
law on national holidays and general holidays: https://www.mevzuat.gov.tr/mevzuat?MevzuatNo=2429&MevzuatTur=1&MevzuatTertip=5 
dates for holidays 'Ramazan Bayramı' and 'Kurban Bayramı': https://vakithesaplama.diyanet.gov.tr/dini_gunler.php
"""


class tr_TR(Locale):
    """
    01-01: [NF] Yılbaşı
    04-23: [NF] Ulusal Egemenlik ve Çocuk Bayramı
    05-01: [NF] Emek ve Dayanışma Günü
    05-19: [NF] Atatürk'ü Anma, Gençlik ve Spor Bayramı
    08-30: [NF] Zafer Bayramı
    10-29: [NF] Cumhuriyet Bayramı
    """

    locale = "tr-TR"

    @staticmethod
    def __ramazan_bayrami_reference(year):
        ramazan_bayrami_reference = {
            2011: SmartDayArrow(2011, 8, 29),
            2012: SmartDayArrow(2012, 8, 18),
            2013: SmartDayArrow(2013, 8,  7),
            2014: SmartDayArrow(2014, 7, 27),
            2015: SmartDayArrow(2015, 7, 16),
            2016: SmartDayArrow(2016, 7,  4),
            2017: SmartDayArrow(2017, 6, 24),
            2018: SmartDayArrow(2018, 6, 14),
            2019: SmartDayArrow(2019, 6,  4),
            2020: SmartDayArrow(2020, 5, 23),
            2021: SmartDayArrow(2021, 5, 12),
            2022: SmartDayArrow(2022, 5,  1),
            2023: SmartDayArrow(2023, 4, 20),
            2024: SmartDayArrow(2024, 4,  9),
            2025: SmartDayArrow(2025, 3, 29),
            2026: SmartDayArrow(2026, 3, 19),
            2027: SmartDayArrow(2027, 3,  8),
        }

        return ramazan_bayrami_reference[year]

    @staticmethod
    def __kurban_bayrami_reference(year):
        kurban_bayrami_reference = {
            2011: SmartDayArrow(2011, 11,  5),
            2012: SmartDayArrow(2012, 10, 24),
            2013: SmartDayArrow(2013, 10, 14),
            2014: SmartDayArrow(2014, 10,  3),
            2015: SmartDayArrow(2015,  9, 23),
            2016: SmartDayArrow(2016,  9, 11),
            2017: SmartDayArrow(2017,  8, 31),
            2018: SmartDayArrow(2018,  8, 20),
            2019: SmartDayArrow(2019,  8, 10),
            2020: SmartDayArrow(2020,  7, 30),
            2021: SmartDayArrow(2021,  7, 19),
            2022: SmartDayArrow(2022,  7,  8),
            2023: SmartDayArrow(2023,  6, 27),
            2024: SmartDayArrow(2024,  6, 15),
            2025: SmartDayArrow(2025,  6,  5),
            2026: SmartDayArrow(2026,  5, 26),
            2027: SmartDayArrow(2027,  5, 15),
        }

        return kurban_bayrami_reference[year]

    def holiday_demokrasi_ve_milli_birlik_gunu(self):
        """
        Democracy and National Unity Day (since 2017)
        07-15 [NF] Demokrasi ve Milli Birlik Günü
        """
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 7, 15),
            "Demokrasi ve Milli Birlik Günü",
            "NF"
        )] if self.year >= 2017 else []

    def holiday_ramazan_bayrami(self):
        """
        Ramazan Bayramı 1.-3. Gün
        """
        reference = self.__ramazan_bayrami_reference(self.year)

        return [Holiday(
            self.locale,
            "",
            reference.shift(days=i),
            "Ramazan Bayramı ({}. Gün)".format(i),
            "NRV"
        ) for i in [1, 2, 3]]

    def holiday_kurban_bayrami(self):
        """
        Kurban Bayramı 1.-4. Gün
        """
        reference = self.__kurban_bayrami_reference(self.year)

        return [Holiday(
            self.locale,
            "",
            reference.shift(days=i),
            "Kurban Bayramı ({}. Gün)".format(i),
            "NRV"
        ) for i in [1, 2, 3, 4]]



# coding=utf-8
from dateutil.easter import EASTER_ORTHODOX

from holidata.utils import SmartDayArrow, easter
from .holidays import Holiday, Locale


class el_GR(Locale):
    """
    01-01: [NF] Πρωτοχρονιά
    01-06: [NRF] Θεοφάνεια
    03-25: [NF] Ευαγγελισμός της Θεοτόκου και Εθνική Ημέρα Ανεξαρτησίας της Ελλάδας
    08-15: [NRF] Κοίμηση της Θεοτόκου
    10-28: [NF] Ημέρα του ΌΧΙ
    12-25: [NRF] Χριστούγεννα
    12-26: [NRF] Επόμενη ημέρα Χριστουγέννων
    48 days before Easter: [NRV] Καθαρά Δευτέρα
    2 days before Easter: [NRV] Μεγάλη Παρασκευή
    1 day before Easter: [NRV] Μεγάλο Σάββατο
    Easter: [NRV] Πάσχα
    1 day after Easter: [NRV] Δευτέρα του Πάσχα
    50 days after Easter: [NRV] Δευτέρα του Αγίου Πνεύματος
    """

    locale = "el-GR"
    easter_type = EASTER_ORTHODOX

    def holiday_may_day(self):
        """
        05-01: [NF] Πρωτομαγιά
        Postponed if it collides with Easter
        """
        date = SmartDayArrow(self.year, 5, 1)
        easter_date = easter(self.year, self.easter_type)

        if date == easter_date:
            date = date.shift(days=2)
        elif date == easter_date.shift(days=1):
            date = date.shift(days=1)

        return [Holiday(
            self.locale,
            "",
            date,
            "Πρωτομαγιά",
            "NF"
        )]



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from holidata.utils import easter, SmartDayArrow
from .holidays import Locale, Holiday

"""
sources:
- Hungarian Constitution - Article J (1)
  https://www.keh.hu/magyarorszag_alaptorvenye/1515-Magyarorszag_Alaptorvenye
- 2017. évi XIII. törvény egyes törvényeknek a nagypéntek munkaszüneti nappá történő nyilvánításával összefüggő módosításáról
  https://mkogy.jogtar.hu/jogszabaly?docid=A1700013.TV
"""


class hu_HU(Locale):
    """
    01-01: [NF] Újév
    03-15: [NF] Az 1848-as forradalom ünnepe
    05-01: [NF] A munka ünnepe
    08-20: [NF] Az államalapítás ünnepe
    10-23: [NF] Az 1956-os forradalom ünnepe
    11-01: [NRF] Mindenszentek
    12-25: [NRF] Karácsony
    12-26: [NRF] Karácsony
    Easter: [NRV] Húsvét
    1 day after Easter: [NRV] Húsvéthétfő
    49 days after Easter: [NRV] Pünkösd
    50 days after Easter: [NRV] Pünkösdhétfő
    """

    locale = "hu-HU"
    easter_type = EASTER_WESTERN

    def holiday_nagypentek(self):
        """
        2 days before Easter: [NRV] Nagypéntek (since 2017)
        """
        if self.year >= 2017:
            return [Holiday(
                self.locale,
                "",
                easter(self.year, self.easter_type).shift(days=-2),
                "Nagypéntek",
                "NRV"
            )]
        else:
            return []

    def holiday_munkaszuneti_nap(self):
        """
        Non-Working days (Munkaszüneti nap)
        When a public holiday falls on a Tuesday or a Thursday, a special decree swaps the preceding Monday or the
        following Friday (respectively) with a not too distant Saturday
        2015: 28/2014. (IX.   24.) NGM rendelet a 2015. évi munkaszüneti napok körüli munkarendről
        2016:  8/2015. (VI.   29.) NGM rendelet a 2016. évi munkaszüneti napok körüli munkarendről
        2018:  9/2017. (V.    19.) NGM rendelet a 2018. évi munkaszüneti napok körüli munkarendről
        2019:  6/2018. (VIII. 23.) PM  rendelet a 2019. évi munkaszüneti napok körüli munkarendről
        2020:  7/2019. (VI.   25.) PM  rendelet a 2020. évi munkaszüneti napok körüli munkarendről
        2021: 14/2020. (V.    13.) ITM rendelet a 2021. évi munkaszüneti napok körüli munkarendről
        2022: 23/2021. (VI.    1.) ITM rendelet a 2022. évi munkaszüneti napok körüli munkarendről
        """
        if self.year == 2015:
            """
            01-02, swapped with 01-10
            08-21, swapped with 08-08
            12-24, swapped with 12-12
            """
            return [
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 1, 2),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2015-01-10 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 8, 21),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2015-08-08 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 12, 24),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2015-12-12 munkanap"
                )]
        if self.year == 2016:
            """
            03-14, swapped with 03-05
            10-31, swapped with 10-15
            """
            return [
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 3, 14),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2016-03-05 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 10, 31),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2016-10-15 munkanap"
                )]
        if self.year == 2018:
            """
            03-16, swapped with 03-10
            04-30, swapped with 04-21
            10-22, swapped with 10-13
            11-02, swapped with 11-10
            12-24, swapped with 12-01
            12-31, swapped with 12-15
            """
            return [
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 3, 16),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2018-03-10 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 4, 30),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2018-04-21 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 10, 22),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2018-10-13 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 11, 2),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2018-11-10 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 12, 24),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2018-12-01 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 12, 31),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2018-12-15 munkanap"
                )]
        if self.year == 2019:
            """
            08-19, swapped with 08-10
            12-24, swapped with 12-07
            12-27, swapped with 12-14
            """
            return [
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 8, 19),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2019-08-10 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 12, 24),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2019-12-07 munkanap"
                ),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 12, 27),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2019-12-14 munkanap"
                )]
        if self.year == 2020:
            """
            08-21, swapped with 08-29
            12-24, swapped with 12-12
            """
            return [
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 8, 21),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2020-08-29 munkanap"),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 12, 24),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2020-12-12 munkanap"
                )]
        if self.year == 2021:
            """
            12-24, swapped with 12-11
            """
            return [
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 12, 24),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2021-12-11 munkanap"
                )]
        if self.year == 2022:
            """
            03-14, swapped with 03-26
            10-31, swapped with 10-15
            """
            return [
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 3, 14),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2022-03-26 munkanap"),
                Holiday(
                    locale=self.locale,
                    region="",
                    date=SmartDayArrow(self.year, 10, 31),
                    description="Munkaszüneti nap",
                    flags="NF",
                    notes="2022-10-15 munkanap"
                )]

        return []



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from .holidays import Locale


class es_US(Locale):
    """
    01-01: [NF] Año Neuvo
    07-04: [NF] Día de la Independiencia
    11-11: [NF] Día de los Veteranos
    12-24: [NRF] Nochebuena
    12-25: [NRF] Navidad
    3. monday in January: [NV] Cumpleaños de Martin Luther King, Jr.
    3. monday in February: [NV] Día del Presidente
    3. monday in April: [MA,ME] [V] Día del Patriota
    1. last monday in May: [NV] Día de los Caídos
    1. monday in September: [NV] Día del Trabajo
    2. monday in October: [NV] Día de Columbus
    4. thursday in November: [NV] Día de Acción de Gracias
    4. friday in November: [NV] Día después de Acción de Gracias
    """

    locale = "es-US"
    easter_type = EASTER_WESTERN



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Locale, Holiday


class hr_HR(Locale):
    """
    01-01: [NF] Nova Godina
    01-06: [NRF] Sveta tri kralja
    05-01: [NF] Praznik rada
    05-30: [NF] Dan državnosti
    06-22: [NF] Dan antifašističke borbe
    08-05: [NF] Dan pobjede i domovinske zahvalnosti i Dan hrvatskih branitelja
    08-15: [NRF] Velika Gospa
    11-01: [NRF] Dan svih svetih
    12-25: [NRF] Božić
    12-26: [NRF] Sveti Stjepan
    Easter: [NRV] Uskrs
    1 day after Easter: [NRV] Uskršnji ponedjeljak
    60 days after Easter: [NRV] Tijelovo
    """

    locale = "hr-HR"
    easter_type = EASTER_WESTERN

    def holiday_dan_sjecanja_na_zrtve_domovinskog_rata_i_dan_sjecanja_na_zrtvu_vukovara_i_skabrnje(self):
        if self.year >= 2020:
            return [Holiday(
                self.locale,
                "",
                SmartDayArrow(self.year, 11, 18),
                "Dan sjećanja na žrtve Domovinskog rata i Dan sjećanja na žrtvu Vukovara i Škabrnje",
                "NF"
            )]
        else:
            return []



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from .holidays import Locale


class pt_PT(Locale):
    """
    01-01: [NF] Ano Novo
    04-25: [NF] Dia da Liberdade
    05-01: [NF] Dia do Trabalhador
    06-10: [NF] Dia de Portugal
    08-15: [NF] Assunção de Nossa Senhora
    10-05: [NF] Implantação da República
    11-01: [NF] Dia de Todos os Santos
    12-01: [NF] Restauração da Independência
    12-08: [NF] Imaculada Conceição
    12-25: [NF] Natal
    47 days before Easter: [NRV] Carnaval
    2 days before Easter: [NRV] Sexta-feira Santa
    Easter: [NRV] Páscoa
    60 days after Easter: [NRV] Corpo de Deus
    """

    locale = "pt-PT"
    easter_type = EASTER_WESTERN



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from .holidays import Locale


class en_US(Locale):
    """
    01-01: [NF] New Year's Day
    07-04: [NF] Independence Day
    11-11: [NF] Veterans Day
    12-24: [NRF] Christmas Eve
    12-25: [NRF] Christmas Day
    3. monday in January: [NV] Birthday of Martin Luther King, Jr.
    3. monday in February: [NV] Washington's Birthday
    3. monday in April: [MA,ME] [V] Patriots' Day
    1. last monday in May: [NV] Memorial Day
    1. monday in September: [NV] Labor Day
    2. monday in October: [NV] Columbus Day
    4. thursday in November: [NV] Thanksgiving Day
    4. friday in November: [NV] Day after Thanksgiving
    """

    locale = "en-US"
    easter_type = EASTER_WESTERN



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Holiday, Locale

#
# Information taken from various government websites around 2020-04-16
#   https://www.canada.ca/fr/emploi-developpement-social/services/normes-travail/rapports/jours-feries.html
#   http://www4.gouv.qc.ca/FR/Portail/Citoyens/Evenements/travailleur-avec-salaire/Pages/jours-feries-chomes-payes.aspx
#   https://www.ontario.ca/document/your-guide-employment-standards-act-0/public-holidays#section-2
#   https://www.cfib-fcei.ca/en/tools-resources/paying-employees-public-holidays-newfoundland-labrador
#   https://www.princeedwardisland.ca/en/information/economic-growth-tourism-and-culture/paid-holidays
#   https://www2.gnb.ca/content/dam/gnb/Departments/petl-epft/PDF/es/FactSheets/PublicHolidaysVacation.pdf
#
# Also those sites for some information
#   https://www.officeholidays.com/holidays/canada/canada-remembrance-day
#   https://www.timeanddate.com/holidays/canada/family-day
#   https://fr.wikipedia.org/wiki/F%C3%AAtes_et_jours_f%C3%A9ri%C3%A9s_au_Canada
#
# I have not checked every province and territory website, but the wikipedia
# summary has been true for everything that I have checked, although it seems
# to be considering more holidays than the bare minimum and also counts holidays
# that are not mandated but usually observed (e.g. St-Patrick's Day in NL which
# is not statutory, but is given to government employees).
#


class en_CA(Locale):
    """
    01-01: [NF] New Year's Day
    06-24: [QC] [F] National Holiday
    07-01: [NF] Canada Day
    11-11: [AB,BC,NB,NL,NT] [F] Remembrance Day
    12-25: [NRF] Christmas Day
    12-26: [NRF] Boxing Day
    2 days before Easter: [NRV] Good Friday
    1 day after Easter: [AB,PE,QC] [RV] Easter Monday
    3. monday in February: [AB,ON,SK,NB] [V] Family Day
    3. monday in February: [MB] [V] Louis Riel Day
    3. monday in February: [PE] [V] Islander Day
    1. monday in August: [NT,NU] [V] August Civic Holiday
    1. monday in August: [SK] [V] Saskatchewan Day
    1. monday in August: [AB,NS] [V] Heritage Day
    1. monday in August: [NB] [V] New Brunswick Day
    1. monday in September: [NV] Labour Day
    2. monday in October: [AB,BC,MB,NL,ON,QC,SK,NT,NU,YT] [V] Thanksgiving Day
    """

    locale = "en-CA"
    easter_type = EASTER_WESTERN

    def holiday_patriot_s_day(self):
        return [Holiday(
                locale=self.locale,
                region='QC',
                date=SmartDayArrow(self.year, 5, 25).shift_to_weekday(
                            'monday', order=1, reverse=True, including=False),
                description="National Patriots' Day",
                flags="V",
                notes="")]


    def holiday_victoria_day(self):
        return [Holiday(
                locale=self.locale,
                region=region,
                date=SmartDayArrow(self.year, 5, 25).shift_to_weekday(
                           'monday', order=1, reverse=True, including=False),
                description="Victoria Day",
                flags="V",
                notes=""
                ) for region in
                ['AB', 'BC', 'MB', 'NS', 'ON', 'SK', 'NT', 'NU', 'YT']
            ]



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from .holidays import Locale


class sk_SK(Locale):
    """
    01-01: [NF] Deň vzniku Slovenskej republiky
    01-06: [NRF] Zjavenie Pána / Traja králi
    05-01: [NF] Sviatok práce
    05-08: [NF] Deň víťazstva nad fašizmom
    07-05: [NRF] Sviatok svätého Cyrila a Metoda
    08-29: [NF] Výročie SNP
    09-01: [NF] Deň Ústavy Slovenskej republiky
    09-15: [NRF] Sedembolestná Panna Mária
    11-01: [NRF] Sviatok všetkých svätých
    11-17: [NF] Deň boja za slobodu a demokraciu
    12-24: [NRF] Štedrý deň
    12-25: [NRF] Prvý sviatok vianočný
    12-26: [NRF] Druhý sviatok vianočný
    2 days before Easter: [NRV] Veľký piatok
    1 day after Easter: [NRV] Veľkonočný pondelok
    """

    locale = "sk-SK"
    easter_type = EASTER_WESTERN



"""
Provides base class for plugins.
"""


class PluginMount(type):
    """
    Metaclass that makes a given class plugin mount.
    All classes inheriting will be referenced in the 'plugins' attribute.
    """

    def __init__(cls, name, bases, attrs):
        super(PluginMount, cls).__init__(name, bases, attrs)

        if not hasattr(cls, 'plugins'):
            cls.plugins = []
        else:
            cls.plugins.append(cls)



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from .holidays import Locale


class fr_FR(Locale):
    """
    01-01: [NF] Jour de l'an
    05-01: [NF] Fête du premier mai
    05-08: [NF] Armistice 1945
    07-14: [NF] Fête nationale
    08-15: [NRF] Assomption
    11-01: [NRF] Toussaint
    11-11: [NF] Armistice 1918
    12-25: [NF] Noël
    1 day after Easter: [NRV] Lundi de Pâques
    39 days after Easter: [NRV] Ascension
    49 days after Easter: [NRV] Pentecôte
    50 days after Easter: [NRV] Lundi de Pentecôte
    """

    locale = "fr-FR"
    easter_type = EASTER_WESTERN



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Holiday, Locale

"""
LEY 51 DE 1983
http://www.suin-juriscol.gov.co/viewDocument.asp?id=1605519
"""

class es_CO(Locale):
    """
    01-01: [NF] Año Nuevo
    05-01: [NF] Día del Trabajo
    07-20: [NF] Grito de Independencia
    08-07: [NF] Batalla de Boyacá
    12-08: [NRF] Inmaculada Concepción
    12-25: [NRF] Navidad
    3 days before Easter: [NRV] Jueves Santo
    2 days before Easter: [NRV] Viernes Santo
    Easter: [NRV] Domingo de Pascua
    43 days after Easter: [NRV] La Ascensión del Señor
    64 days after Easter: [NRV] Corpus Christi
    71 days after Easter: [NRV] El Sagrado Corazón de Jesús
    """

    locale = "es-CO"
    easter_type = EASTER_WESTERN

    def holiday_reyes(self):
        """First Monday after January 6."""
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 1, 6).shift_to_weekday('monday', including=True),
            "Día de los Reyes Magos",
            "NRV"
        )]

    def holiday_san_jose(self):
        """First Monday after March 19."""
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 3, 19).shift_to_weekday('monday', including=True),
            "Día de San José",
            "NRV"
        )]

    def holiday_san_pedro_san_pablo(self):
        """First Monday after June 29."""
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 6, 29).shift_to_weekday('monday', including=True),
            "San Pedro y San Pablo",
            "NRV"
        )]

    def holiday_asuncion(self):
        """First Monday after August 15."""
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 8, 15).shift_to_weekday('monday', including=True),
            "Asunción de la Virgen",
            "NRV"
        )]

    def holiday_dia_raza(self):
        """First Monday after October 12."""
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 10, 12).shift_to_weekday('monday', including=True),
            "Día de la Raza",
            "NV"
        )]

    def holiday_todos_santos(self):
        """First Monday after November 1."""
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 11, 1).shift_to_weekday('monday', including=True),
            "Todos los Santos",
            "NRV"
        )]

    def holiday_independencia_cartagena(self):
        """First Monday after November 11."""
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 11, 11).shift_to_weekday('monday', including=True),
            "Independencia de Cartagena",
            "NV"
        )]




from .holidays import Country


class BR(Country):
    id = "BR"
    languages = ["pt"]
    default_lang = "pt"
    regions = ["AC", "AL", "AP", "AM", "BA", "CE", "DF", "ES", "GO", "MA", "MT", "MS", "MG", "PA", "PB", "PR", "PE",
               "PI", "RJ", "RN", "RS", "RO", "RR", "SC", "SP", "SE", "TO"]



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Locale, Holiday

"""
source: https://almanakka.helsinki.fi/en/flag-days-and-holidays-in-finland.html
"""


class sv_FI(Locale):
    """
    01-01: [NF] Nyårsdagen
    01-06: [NRF] Trettondedagen
    05-01: [NF] Första maj
    12-06: [NF] Självständighetsdagen
    12-25: [NRF] Juldagen
    12-26: [NRF] Annandag jul
    2 days before Easter: [NRV] Långfredagen
    Easter: [NRV] Påskdagen
    1 day after Easter: [NRV] Annandag påsk
    39 days after Easter: [NRV] Kristi himmelfärdsdag
    49 days after Easter: [NRV] Pingst
    """

    locale = "sv-FI"
    easter_type = EASTER_WESTERN

    def holiday_midsommardagen(self):
        """
        Saturday between 20 and 26 June: Midsommardagen
        """
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 6, 19).shift_to_weekday('saturday', order=1, reverse=False),
            "Midsommardagen",
            "NRV"
        )]

    def holiday_alla_helgons_dag(self):
        """
        Saturday between 31 October and 6 November: Alla helgons dag (All Saints' Day)
        """
        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 10, 30).shift_to_weekday('saturday', order=1, reverse=False),
            "Alla helgons dag",
            "NRV"
        )]



import csv
import io
import json

from holidata.plugin import PluginMount


class Emitter(object, metaclass=PluginMount):
    type = None

    def __init__(self):
        if self.type is None:
            raise ValueError("Emitter {0} does not provide its type".format(self.__class__.__name__))

    def output(self, locale):
        pass


class JsonEmitter(Emitter):
    type = "json"

    def output(self, locale):
        export_data = [h.as_dict() for h in locale.holidays]
        export_data.sort(key=lambda x: x['date'])
        return "\n".join([json.dumps(h, ensure_ascii=False, sort_keys=False, indent=None, separators=(',', ':')) for h in export_data]) + "\n"


class CsvEmitter(Emitter):
    type = "csv"

    def output(self, locale):
        export_data = [h.as_dict() for h in locale.holidays]
        export_data.sort(key=lambda x: x['date'])
        result = io.StringIO()

        writer = csv.DictWriter(result,
                                ["locale", "region", "date", "description", "type", "notes"],
                                quoting=csv.QUOTE_ALL,
                                lineterminator='\n')
        writer.writeheader()
        writer.writerows(export_data)

        return result.getvalue()


class YamlEmitter(Emitter):
    type = "yaml"

    def output(self, locale):
        export_data = [h.as_dict() for h in locale.holidays]
        export_data.sort(key=lambda x: x["date"])

        output = "%YAML 1.1\n"
        output += "---\n"
        for holiday in export_data:
            output += "  holiday:\n"

            for key in ["locale", "region", "date", "description", "type", "notes"]:
                value = holiday[key]

                if value is not None and value != "":
                    output += "    {}: {}\n".format(key, value)
                else:
                    output += "    {}:\n".format(key)

        output += "...\n"
        return output


class XmlEmitter(Emitter):
    type = "xml"

    def output(self, locale):
        export_data = [h.as_dict() for h in locale.holidays]
        export_data.sort(key=lambda x: x["date"])

        output = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n"
        output += "<holidays>\n"

        for holiday in export_data:
            output += "  <holiday>\n"

            for key in ["locale", "region", "date", "description", "type", "notes"]:
                value = holiday[key] if key in holiday else ""
                output += "    <{0}>{1}</{0}>\n".format(key, value if value is not None else "")

            output += "  </holiday>\n"

        output += "</holidays>\n"
        return output



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from holidata.utils import easter, SmartDayArrow
from .holidays import Locale, Holiday

"""
source: https://www.zakonyprolidi.cz/cs/2000-245, §1 and §2.
        https://www.zakonyprolidi.cz/cs/2000-245/zneni-20200201 (>2020-02-01)
        https://www.zakonyprolidi.cz/cs/2000-245/zneni-20190401 (>2019-04-01)
        https://www.zakonyprolidi.cz/cs/2000-245/zneni-0 (>2000-08-09)
"""


class cs_CZ(Locale):
    """
    01-01: [NF] Nový rok
    01-01: [NF] Den obnovy samostatného českého státu
    05-01: [NF] Svátek práce
    05-08: [NF] Den vítězství
    07-05: [NRF] Den slovanských věrozvěstů Cyrila a Metoděje
    07-06: [NRF] Den upálení mistra Jana Husa
    09-28: [NRF] Den české státnosti
    10-28: [NF] Den vzniku samostatného československého státu
    12-24: [NRF] Štědrý den
    12-25: [NRF] 1. svátek vánoční
    12-26: [NRF] 2. svátek vánoční
    1 day after Easter: [NRV] Velikonoční pondělí
    """

    locale = "cs-CZ"
    easter_type = EASTER_WESTERN

    def holiday_velky_patek(self):
        """
        2 days before Easter: [NRV] Velký pátek
        since 2016
        """
        if self.year >= 2016:
            return [Holiday(
                self.locale,
                "",
                easter(self.year, self.easter_type).shift(days=-2),
                "Velký pátek",
                "NRV"
            )]

        return []

    def holiday_den_boje_za_svobodu_a_demokracii_a_mezinarodni_den_studentstva(self):
        """
        11-17: [NF]
        before 2019-04-01: Den boje za svobodu a demokracii
        before 2019-04-01: Den boje za svobodu a demokracii a Mezinárodní den studentstva
        """
        if self.year < 2019:
            name = "Den boje za svobodu a demokracii"
        else:
            name = "Den boje za svobodu a demokracii a Mezinárodní den studentstva"

        return [Holiday(
            self.locale,
            "",
            SmartDayArrow(self.year, 11, 17),
            name,
            "NF"
        )]



# coding=utf-8
from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow, month_reference
from .holidays import Holiday, Locale


class en_GB(Locale):
    """
    01-01: [NF] New Year's Day
    12-25: [NRF] Christmas Day
    12-26: [NF] Boxing Day
    2 days before Easter: [NRV] Good Friday
    1 day after Easter: [NRV] Easter Monday
    1. monday in may: [NV] Early May Bank Holiday
    1. last monday in august: [NV] August Bank Holiday
    """

    locale = "en-GB"
    easter_type = EASTER_WESTERN

    def holiday_new_years_day_observed(self):
        date = SmartDayArrow(self.year, 1, 1)

        if date.weekday() in ['saturday', 'sunday']:
            return [Holiday(
                    locale=self.locale,
                    region="",
                    date=date.shift_to_weekday('monday', including=True),
                    description="New Year's Day (observed)",
                    flags="NV",
                    notes="")]

        return []

    def holiday_spring_bank_holiday(self):
        """
        1. last monday in may: [NV] Spring Bank Holiday
        2012: Moved to June 4, because of Queen’s Diamond Jubilee
        2022: Moved to June 2, because of Queen's Platinum Jubilee
        """
        if self.year == 2012:
            date = SmartDayArrow(self.year, 6, 4)
        elif self.year == 2022:
            date = SmartDayArrow(self.year, 6, 2)
        else:
            date = month_reference(self.year,
                                   "may",
                                   first=False) \
                .shift_to_weekday("monday",
                                  order=1,
                                  reverse=True,
                                  including=True)

        return [Holiday(
            locale=self.locale,
            region="",
            date=date,
            description="Spring Bank Holiday",
            flags="NV",
            notes="")]

    def holiday_christmas_day_observed(self):
        date = SmartDayArrow(self.year, 12, 25)

        if date.weekday() == 'saturday':
            return [Holiday(
                    locale=self.locale,
                    region="",
                    date=date.shift_to_weekday('monday', including=True),
                    description="Christmas Day (observed)",
                    flags="NV",
                    notes="")]

        elif date.weekday() == 'sunday':
            return [Holiday(
                    locale=self.locale,
                    region="",
                    date=date.shift_to_weekday('tuesday', including=True),
                    description="Christmas Day (observed)",
                    flags="NV",
                    notes="")]

        return []

    def holiday_boxing_day_observed(self):
        date = SmartDayArrow(self.year, 12, 26)

        if date.weekday() == 'sunday':
            return [Holiday(
                    locale=self.locale,
                    region="",
                    date=date.shift_to_weekday('tuesday', including=True),
                    description="Boxing Day (observed)",
                    flags="NV",
                    notes="")]

        elif date.weekday() == 'saturday':
            return [Holiday(
                    locale=self.locale,
                    region="",
                    date=date.shift_to_weekday('monday', including=True),
                    description="Boxing Day (observed)",
                    flags="NV",
                    notes="")]

        return []

    def holiday_royal_jubilees(self):
        """
        2012-06-05: Queen's Diamond Jubilee
        2022-06-03: Queen's Platinum Jubilee
        """
        if self.year == 2012:
            return [Holiday(
                locale=self.locale,
                region="",
                date=SmartDayArrow(self.year, 6, 5),
                description="Queen's Diamond Jubilee",
                flags="NV",
                notes="")]

        if self.year == 2022:
            return [Holiday(
                locale=self.locale,
                region="",
                date=SmartDayArrow(self.year, 6, 3),
                description="Queen's Platinum Jubilee",
                flags="NV",
                notes="")]

        return []



import re

from holidata.plugin import PluginMount
from holidata.utils import SmartDayArrow, month_reference, easter


class Holiday(object):
    """
    A sheer container for one holiday.
    """
    def __init__(self, locale, region, date, description, flags="", notes=""):
        self.locale = locale
        self.region = region
        self.date = date
        self.description = description
        self.flags = flags
        self.notes = notes

    def as_dict(self):
        return {
            'locale': self.locale,
            'region': self.region,
            'date': self.date.strftime('%Y-%m-%d'),
            'description': self.description,
            'type': self.flags,
            'notes': self.notes
        }


class Country(object, metaclass=PluginMount):
    """
    Represents holidays of a given country
    """
    id = None
    languages = []
    default_lang = None
    easter_type = None

    def __init__(self):
        if self.id is None:
            raise ValueError("Country '{0}' does not provide its id!".format(self.__class__.__name__))

        if not self.languages:
            raise ValueError("Country '{0}' does not list languages!".format(self.__class__.__name__))

        if self.default_lang is not None and self.default_lang not in self.languages:
            raise ValueError("Country '{0}' does not list language '{1}'!".format(self.__class__.__name__, self.default_lang))


class Locale(object, metaclass=PluginMount):
    """
    Represents holidays in a given locale.
    """
    locale = None
    easter_type = None

    fixed_regex = re.compile(r'^\s*(?P<month>\d\d)-(?P<day>\d\d): '
                             r'(\[(?P<regions>[^]]+)\]\s+)?'
                             r'\[(?P<flags>[A-Z]*)\] (?P<description>.*)$', re.UNICODE)
    nth_weekday_regex = re.compile(r'^\s*(?P<order>\d+)\.(?P<last> last | )'
                                   r'(?P<weekday>[a-z]+) in (?P<month>[a-zA-Z]+):\s+'
                                   r'(\[(?P<regions>[^]]+)\]\s+)?'
                                   r'\[(?P<flags>[A-Z]*)\] (?P<description>.*)$', re.UNICODE)
    easter_shift_regex = re.compile(r'^\s*((?P<days>\d+) day(s)? (?P<direction>(before|after)) )?Easter:\s+'
                                    r'(\[(?P<regions>[^]]+)\]\s+)?'
                                    r'\[(?P<flags>[A-Z]*)\] (?P<description>.*)$', re.UNICODE)

    def __init__(self, year):
        if self.locale is None:
            raise ValueError("Locale {0} does not provide its locale".format(self.__class__.__name__))

        self.year = year

    @property
    def holidays(self):
        """
        Yield all the Holiday objects corresponding to the definitions in the
        self.__doc__ and also as given by the dynamic self.holiday_* methods.
        """
        # First process lines in the __doc__
        for line in self.__doc__.splitlines():
            # Skip empty lines
            if not line.strip():
                continue

            holidata = self._parse_holidata(line)

            if holidata is None:
                print("Following line could not be processed: '{}'".format(line))
                continue

            for region in holidata['regions']:
                yield Holiday(
                    locale=self.locale,
                    region=region,
                    date=holidata['date'],
                    description=holidata['description'],
                    flags=holidata['flags'],
                    notes=holidata['notes'],
                )

        # Second, call holiday functions
        for method in [getattr(self, func) for func in dir(self) if func.startswith('holiday_')]:
            holidays = method()

            for holiday in holidays:
                yield holiday

    def _parse_holidata(self, line):
        function_map = [
            (self.fixed_regex, self._date_from_fixed_reference),
            (self.nth_weekday_regex, self._date_from_weekday_reference),
            (self.easter_shift_regex, self._date_from_easter_reference),
        ]

        for reg_exp, create_date_from in function_map:
            m = reg_exp.search(line)
            if m is not None:
                return dict(regions=([x.strip() for x in m.group('regions').split(',')] if m.group('regions') is not None else [""]),
                            date=create_date_from(m),
                            description=m.group('description'),
                            flags=m.group('flags'),
                            notes="")

        return None

    def _date_from_fixed_reference(self, m):
        return SmartDayArrow(self.year, int(m.group('month')), int(m.group('day')))

    def _date_from_weekday_reference(self, m):
        return month_reference(self.year,
                               m.group('month'),
                               first=m.group('last').strip() == '') \
            .shift_to_weekday(m.group('weekday'),
                              order=int(m.group('order')),
                              reverse=m.group('last').strip() == 'last',
                              including=True)

    def _date_from_easter_reference(self, m):
        if self.easter_type is None:
            raise ValueError("Locale {0} does not provide its easter type (WESTERN|ORTHODOX)".format(self.__class__.__name__))

        return easter(self.year, self.easter_type) \
            .shift(days=int((m.group('days')) if m.group('days') is not None else 0) * (1 if m.group('direction') == 'after' else -1))
